/*
 * Example program to illustrate how to read random data from
 * the Alea I on Linux, NetBSD, and other systems using libusb.
 *
 * (c) Copyright 2004-2012 Araneus Information Systems Oy.
 * All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include <fcntl.h>

#include <usb.h>

#define USB_VENDOR_ARANEUS		0x12d8
#define USB_ARANEUS_PRODUCT_ALEA	0x0001

/* Number of bytes to read at a time.  Must be a multiple of 64. */
#define BUFFER_SIZE 8192

/* Timeout in milliseconds */
#define TIMEOUT 5000

static void
fatal(const char *format, ...) {
    va_list args;
    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);
    fprintf(stderr, "\n");
    exit(1);
}

static void
usage(void) {
    fatal("usage: randomfile [-b] [length]");
}

int
main(int argc, char *argv[]) {
    int nbytes;		/* Number of bytes to read, or -1 if no limit */
    int binary = 0;
    int r;
    unsigned char buffer[BUFFER_SIZE];
    unsigned int remaining;
    struct usb_bus *bus;
    struct usb_device *dev;
    struct usb_dev_handle *handle;

    argv++, argc--;
    while (argc && argv[0][0] == '-') {
	if (strcmp(argv[0], "-b") == 0) {
	    binary++;
	} else {
	    usage();
	}
	argv++, argc--;
    }

    if (argc == 0)
	nbytes = -1;
    else if (argc == 1)
	nbytes = atoi(argv[0]);
    else
	usage();

    usb_init();
    usb_find_busses();
    usb_find_devices();

    for (bus = usb_busses; bus != NULL; bus = bus->next) {
	for (dev = bus->devices; dev != NULL; dev = dev->next) {
	    if (dev->descriptor.idVendor == USB_VENDOR_ARANEUS &&
		dev->descriptor.idProduct == USB_ARANEUS_PRODUCT_ALEA)
		goto found;
	}
    }
    fatal("No Alea device found");

 found:
    handle = usb_open(dev);
    r = usb_claim_interface(handle, 0);
    if (r < 0)
	fatal("unable to claim Alea interface 0");

    remaining = nbytes == -1 ? BUFFER_SIZE : nbytes;
    while (remaining > 0) {
	int nreq;
	int nreq_rounded;
	int nread;
	int i;

	nreq = remaining <= BUFFER_SIZE ? remaining : BUFFER_SIZE;
	/* Round the size up to the nearest multiple of 64 bytes */
	nreq_rounded = (nreq + 63) & ~63;

	nread = usb_bulk_read(handle, 1, (char *) buffer, nreq_rounded,
			      TIMEOUT);
	if (nread < 0)
	    fatal("error reading from Alea device");
	if (nread != nreq_rounded)
	    fatal("short read reading from Alea device");
	for (i = 0; i < nreq; i++) {
	    if (binary)
		putchar(buffer[i]);
	    else
		printf("%02x", buffer[i]);
	}
	fflush(stdout);
	if (nbytes != -1)
	    remaining -= nreq;
    }
    usb_release_interface(handle, 0);
    usb_close(handle);
    return 0;
}
