//
// Interface to Alea I device driver
//
// (c) Copyright 2004-2012 Araneus Information Systems Oy.
// All rights reserved.
//

#include <windows.h>
#include <setupapi.h>
#include <initguid.h>
#include <winioctl.h>

#include <assert.h>
#include <stdlib.h>

#define MAX_DEVPATH_LENGTH 256

#include "alea.h"

static char ident[] = "$RCSfile: alea.cpp,v $ $Revision: 1.24 $";

/////////////////////////////////////////////////////////////////////////////
//
// Common utility code
//

static bool
get_device_path(LPGUID InterfaceGuid, char *path_buf, size_t path_buf_len,
		unsigned int n)
{
    // Retrieve device list for the specified GUID.
    HDEVINFO devices =
	SetupDiGetClassDevs(InterfaceGuid, NULL, NULL,
			    DIGCF_PRESENT | DIGCF_DEVICEINTERFACE);
    if (devices == NULL) {
	SetLastError(ALEA_ERROR_NODEV);
	return false;
    }

    PSP_DEVICE_INTERFACE_DETAIL_DATA device_detail = NULL;

    for (int i = 0; i <= 127; i++) {
	SP_DEVICE_INTERFACE_DATA device_info;
	ZeroMemory(&device_info, sizeof(device_info));
	device_info.cbSize = sizeof(SP_DEVICE_INTERFACE_DATA);
	
	// Retrieve a context structure for a device interface of a
	// device information set.
	if (! SetupDiEnumDeviceInterfaces(devices, 0,
					  InterfaceGuid,
					  i, &device_info))
	{
	    if (GetLastError() == ERROR_NO_MORE_ITEMS) {
		SetLastError(ALEA_ERROR_NODEV);
		goto cleanup;
	    } else {
		continue;
	    }
	}
	// Determine the buffer size required for
	// SetupDiGetDeviceInterfaceDetail().
	ULONG len = 0;
	SetupDiGetDeviceInterfaceDetail(devices,
					&device_info,
					NULL,
					0,
					&len,
					NULL);
	device_detail =
	    (PSP_DEVICE_INTERFACE_DETAIL_DATA) malloc(len);
	if (! device_detail) {
	    SetLastError(ERROR_OUTOFMEMORY);
	    goto cleanup;
	}
	    
	device_detail->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);

	// Retrieve detailed device interface information.
	ULONG dummy;
	if (! SetupDiGetDeviceInterfaceDetail(devices,
					      &device_info,
					      device_detail,
					      len,
					      &dummy,
					      NULL))
	{
	    free(device_detail);
	    SetLastError(ALEA_ERROR_NODEV);
	    goto cleanup;
	}

	if (n-- > 0) {
	    free(device_detail);
	    continue;
	}

	// Success
	break;
    }

    if (strlen(device_detail->DevicePath) + 1  > path_buf_len) {
	free(device_detail);
	SetLastError(ERROR_BUFFER_OVERFLOW);
	goto cleanup;
    }

    strcpy(path_buf, device_detail->DevicePath);
    free(device_detail);
    SetupDiDestroyDeviceInfoList(devices);
    return true;

 cleanup:
    SetupDiDestroyDeviceInfoList(devices);
    return false;
}

/////////////////////////////////////////////////////////////////////////////
//
// Interface to the legacy Windows XP driver
//

// {3599c85e-eff0-4dc0-9833-6c5aaf4a0d54}
DEFINE_GUID(GUID_INTERFACE_ALEA,
	    0x3599c85e,
	    0xeff0,
	    0x4dc0,
	    0x98, 0x33, 0x6c, 0x5a, 0xaf, 0x4a, 0x0d, 0x54);

#define ALEA_IOCTL_INDEX             0x0800

#define IOCTL_ALEA_RESET_DEVICE          CTL_CODE(FILE_DEVICE_UNKNOWN,  \
                                                  ALEA_IOCTL_INDEX + 1, \
                                                  METHOD_BUFFERED,      \
                                                  FILE_ANY_ACCESS)

#define IOCTL_ALEA_RESET_PIPE            CTL_CODE(FILE_DEVICE_UNKNOWN,  \
                                                  ALEA_IOCTL_INDEX + 2, \
                                                  METHOD_BUFFERED,      \
                                                  FILE_ANY_ACCESS)

class alea_xp_t: public alea_t {
public:
    alea_xp_t(HANDLE h): filehandle(h) { }
    ~alea_xp_t();
private:
    bool block_read(unsigned char *buffer, unsigned int nbytes);
    bool get_serial(char *buf, size_t bufsize) { return false; }

    HANDLE filehandle;
};

alea_t *
alea_open_xp(unsigned int n) {
    char device_path[MAX_DEVPATH_LENGTH];
    static char pipe_subpath[] = "\\PIPE00";
    
    bool rb = get_device_path((LPGUID) &GUID_INTERFACE_ALEA,
			     device_path,
			     sizeof(device_path) - strlen(pipe_subpath),
			     n);
    if (! rb)
	return 0;
    strcat(device_path, pipe_subpath);

    HANDLE filehandle = 
	CreateFile(device_path,
		   GENERIC_WRITE | GENERIC_READ,
		   FILE_SHARE_WRITE | FILE_SHARE_READ,
		   NULL,
		   OPEN_EXISTING,
		   0,
		   NULL);
    if (filehandle == INVALID_HANDLE_VALUE)
	return 0;

    unsigned long dummy;
    int r;
    r = DeviceIoControl(filehandle,
			IOCTL_ALEA_RESET_PIPE,
			NULL, 0,
			NULL, 0,
			&dummy,
			NULL);
    if (! r) {
	CloseHandle(filehandle);
	return 0;
    }

    return new alea_xp_t(filehandle);
}

alea_xp_t::~alea_xp_t() {
    CloseHandle(filehandle);
}

bool
alea_xp_t::block_read(unsigned char *buffer, unsigned int nbytes) {
    assert(nbytes % ALEA_USB_BLOCKSIZE == 0);
    ULONG nread;
    bool r = ReadFile(filehandle, buffer, nbytes, &nread, NULL);
    return (r && nread == nbytes);
}

/////////////////////////////////////////////////////////////////////////////
//
// Interface to the WinUSB driver
//

// {93c78847-0a8c-477a-8878-afbec1d5145a}
DEFINE_GUID(GUID_INTERFACE_ALEA_WINUSB,
	    0x93c78847,
	    0x0a8c,
	    0x477a,
            0x88, 0x78, 0xaf, 0xbe, 0xc1, 0xd5, 0x14, 0x5a);

// We don't include winusb.h, because it's only shipped with the DDK
// and will not compile cleanly in a user-mode build environment.

typedef PVOID WINUSB_INTERFACE_HANDLE, *PWINUSB_INTERFACE_HANDLE;

#define WINUSB_FUNC_DECL(ret, name, args) \
    typedef ret (__stdcall * ##name##_t) args; \
    ##name##_t name

WINUSB_FUNC_DECL(BOOL, WinUsb_Initialize, (
    IN HANDLE DeviceHandle,
    OUT PWINUSB_INTERFACE_HANDLE InterfaceHandle
    ));

WINUSB_FUNC_DECL(BOOL, WinUsb_Free, (
    IN  WINUSB_INTERFACE_HANDLE InterfaceHandle
    ));

WINUSB_FUNC_DECL(BOOL, WinUsb_ResetPipe, (
    IN  WINUSB_INTERFACE_HANDLE InterfaceHandle,
    IN  UCHAR PipeID
    ));

WINUSB_FUNC_DECL(BOOL, WinUsb_ReadPipe, (
    IN  WINUSB_INTERFACE_HANDLE InterfaceHandle,
    IN  UCHAR PipeID,
    IN  PUCHAR Buffer,
    IN  ULONG BufferLength,
    OUT PULONG LengthTransferred,
    IN  LPOVERLAPPED Overlapped
    ));

WINUSB_FUNC_DECL(BOOL, WinUsb_GetDescriptor, (
    IN WINUSB_INTERFACE_HANDLE InterfaceHandle,
    IN UCHAR  DescriptorType,
    IN UCHAR  Index,
    IN USHORT  LanguageID,
    OUT PUCHAR  Buffer,
    IN ULONG  BufferLength,
    OUT PULONG  LengthTransferred
    ));

typedef struct _USB_DEVICE_DESCRIPTOR { 
    UCHAR bLength;
    UCHAR bDescriptorType;
    USHORT bcdUSB;
    UCHAR bDeviceClass;
    UCHAR bDeviceSubClass;
    UCHAR bDeviceProtocol;
    UCHAR bMaxPacketSize0;
    USHORT idVendor;
    USHORT idProduct;
    USHORT bcdDevice;
    UCHAR iManufacturer;
    UCHAR iProduct;
    UCHAR iSerialNumber;
    UCHAR bNumConfigurations;
} USB_DEVICE_DESCRIPTOR;

typedef struct _USB_STRING_DESCRIPTOR {
    UCHAR  bLength;
    UCHAR  bDescriptorType;
    WCHAR  bString[1];
} USB_STRING_DESCRIPTOR;

#define USB_DEVICE_DESCRIPTOR_TYPE 1
#define USB_STRING_DESCRIPTOR_TYPE 3

class alea_winusb_t: public alea_t {
public:
    alea_winusb_t(HMODULE d, HANDLE f, WINUSB_INTERFACE_HANDLE h):
        dll(d), filehandle(f), usb_handle(h) { }
    ~alea_winusb_t();
private:
    bool block_read(unsigned char *buffer, unsigned int nbytes);
    bool get_serial(char *buf, size_t bufsize);

    HMODULE dll;
    HANDLE filehandle;
    WINUSB_INTERFACE_HANDLE usb_handle;
};

HMODULE
load_winusb() {
    HMODULE dll;
    const char dllname[] = "winusb.dll";
    dll = GetModuleHandle(dllname);
    if (! dll)
	dll = LoadLibrary(dllname);
    if (! dll)
	return 0;

#define WINUSB_FUNC_GET(name) \
    do { \
	name = (##name##_t) GetProcAddress(dll, #name); \
        if (! name) \
            return 0; \
    } while(0) 
    
    WINUSB_FUNC_GET(WinUsb_Initialize);
    WINUSB_FUNC_GET(WinUsb_Free);
    WINUSB_FUNC_GET(WinUsb_ResetPipe);
    WINUSB_FUNC_GET(WinUsb_ReadPipe);
    WINUSB_FUNC_GET(WinUsb_GetDescriptor);

    return dll;
}

alea_t *
alea_open_winusb(unsigned int n) {
    HMODULE dll = load_winusb();
    if (! dll)
	goto cleanup4;

    char device_path[MAX_DEVPATH_LENGTH];
    bool r = get_device_path((LPGUID) &GUID_INTERFACE_ALEA_WINUSB,
			     device_path,
			     sizeof(device_path),
			     n);
    if (! r)
	goto cleanup3;

    HANDLE filehandle =
	CreateFile(device_path,
		   GENERIC_WRITE | GENERIC_READ,
		   FILE_SHARE_WRITE | FILE_SHARE_READ,
		   NULL,
		   OPEN_EXISTING,
		   FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED,
		   NULL);
    if (filehandle == INVALID_HANDLE_VALUE)
	goto cleanup3;

    WINUSB_INTERFACE_HANDLE usb_handle;
    BOOL result = WinUsb_Initialize(filehandle, &usb_handle);
    if (! result)
	goto cleanup2;

    result = WinUsb_ResetPipe(usb_handle, 0x81);
    if (! result)
        goto cleanup1;

    return new alea_winusb_t(dll, filehandle, usb_handle);

 cleanup1:
    WinUsb_Free(usb_handle);
 cleanup2:
    CloseHandle(filehandle);
 cleanup3:
    FreeLibrary(dll);
 cleanup4:
    return 0;
}

alea_winusb_t::~alea_winusb_t() {
    WinUsb_Free(usb_handle);
    CloseHandle(filehandle);
    FreeLibrary(dll);
}

bool
alea_winusb_t::block_read(unsigned char *buffer, unsigned int nbytes) {
    assert(nbytes % ALEA_USB_BLOCKSIZE == 0);
    ULONG nread;
    BOOL r = WinUsb_ReadPipe(usb_handle,
			     0x81,
			     buffer,
			     nbytes,
			     &nread,
			     NULL);
    return (r && nread == nbytes);
}

bool
alea_winusb_t::get_serial(char *buf, size_t bufsize) {
    unsigned long bytes_transferred;
    USB_DEVICE_DESCRIPTOR dd;
    bool r = WinUsb_GetDescriptor(usb_handle,
				  USB_DEVICE_DESCRIPTOR_TYPE,
				  0, 0,
				  (unsigned char *) &dd, sizeof(dd),
				  &bytes_transferred);
    if (! r)
	return false;

    union {
	USB_STRING_DESCRIPTOR sdesc;
	unsigned char buf[512];
    } u;
    
    r = WinUsb_GetDescriptor(usb_handle,
			     USB_STRING_DESCRIPTOR_TYPE,
			     dd.iSerialNumber, 0,
			     u.buf, sizeof(u),
			     &bytes_transferred);
    if (! r || bytes_transferred < 2 || bytes_transferred >= sizeof(u))
	return false;

    // Calculate the number of characters in the string.  The string
    // descriptor length consists of a two-byte header and two bytes
    // per character.
    int nchars = (u.sdesc.bLength - 2) / 2;

    if (bufsize < nchars + 1)
	return false;

    int i;
    for (i = 0; i < nchars; i++) {
	buf[i] = (char) u.sdesc.bString[i];
    }
    buf[i] = '\0';
    return true;
}

/////////////////////////////////////////////////////////////////////////////

bool
alea_t::read(unsigned char *dst, size_t nbytes) {
    size_t remain = nbytes;
    
    while (remain) {
	size_t chunk = remain;
	if (chunk > ALEA_BUFFERSIZE)
	    chunk = ALEA_BUFFERSIZE;
	// We need "chunk" bytes of randomness, and
	// we already have "nbuffered" bytes
	size_t read_chunk;
	if (nbuffered < chunk) {
	    size_t needed = chunk - nbuffered;
	    // We always read whole ALEA_USB_BLOCKSIZE blocks, so round up
	    read_chunk = (needed + (ALEA_USB_BLOCKSIZE - 1)) & ~(ALEA_USB_BLOCKSIZE - 1);
	    bool r = block_read(buffer + nbuffered, (unsigned int) read_chunk);
	    if (! r)
		return false;
	} else {
	    read_chunk = 0;
	}
	// Copyout
	memcpy(dst, buffer, chunk);
	dst += chunk;
	remain -= chunk;
	// Move leftover data to the beginning of the buffer
	nbuffered = (unsigned int) (nbuffered + read_chunk - chunk);
	if (nbuffered)
	    memmove(buffer, buffer + chunk, nbuffered);
    }
    return true;
}

// Virtual constructors

alea_t *
alea_open_n(unsigned int n) {
    // Try each driver in turn
    alea_t *alea = alea_open_winusb(n);
    if (alea)
      return alea;
    alea = alea_open_xp(n);
    if (alea)
      return alea;
    // error code was set by one of the above functions
    return 0;
}

alea_t *
alea_open(void) {
    return alea_open_n(0);
}
