//
// Interface to Alea I device driver
//
// (c) Copyright 2004-2012 Araneus Information Systems Oy.
// All rights reserved.
//
//
// To obtain random bits from the Alea I device, use code like
// the following.  Details of error handling have been omitted.
//
//    alea_t *alea = alea_open();
//    if (! alea)
//        error(...);
//    bool success = alea->read(buffer, nbytes);
//    if (! success)
//        error(...);
//    delete alea; // this closes the device
//

// The maximum number of bytes we will read from the Alea I in
// a single system call
const unsigned int ALEA_BUFFERSIZE = 8192;

// Reads from the Alea I will be in multiples of this block size;
// it must be a power of two, and must be at least 64 bytes so that
// we read whole USB blocks.
const unsigned int ALEA_USB_BLOCKSIZE = 128;

// Abstract base class interfacing to the Alea I.  To construct an
// object of this class, call either alea_open() or alea_open_n().

class alea_t {
public:
    alea_t(): nbuffered(0) { }

    // Clean up, closing the device.
    virtual ~alea_t() { }

    // Return "nbytes" bytes of random data into "buffer".
    // Returns true iff successful.
    bool read(unsigned char *buffer, size_t nbytes);

    // Get the serial number of the device.  This operation is only
    // supported by device driver version 2.0 or newer; it will fail
    // (returning false) if the device is using the 1.x driver.  The
    // size of the buffer is given by "bufsize", and should be at
    // least 7 to fit a six digit serial number and the terminating
    // null.
    virtual bool get_serial(char *buf, size_t bufsize) = 0;

private:
    // Read in multiples of 64 bytes (whole USB packets)
    virtual bool block_read(unsigned char *buffer, unsigned int nbytes) = 0;

    // Buffer for reading from USB and storing residual data.
    // Invariant: the first "nbuffered" bytes of "buffer" contain
    // valid buffered randomness.  The last byte of the buffer is unused.
    unsigned char buffer[ALEA_BUFFERSIZE+ALEA_USB_BLOCKSIZE];
    unsigned int nbuffered;
};

// Find the first Alea I device connected to the system and return a
// pointer to an alea_t object.  In case of error, a null pointer
// is returned, and an error code can be retrieved by calling GetLastError().
// If the cause of the error is that no Alea I device could be found,
// the error code will be ALEA_ERROR_NODEV.

alea_t *alea_open();

// As above, but find the Nth device (counting from 0), to support 
// the use of multiple connected devices.

alea_t *alea_open_n(unsigned int);

#define ALEA_ERROR_NODEV (0x20000001)
